/*
 *	Copyright (C) 2009 Jonathan Naylor, G4KLX
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; version 2 of the License.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 */

#include "DStarRadioHeaderProcessor.h"

#include "DStarFEC.h"
#include "DStarInterleaver.h"
#include "CCITTChecksum.h"

#define	_CRT_SECURE_NO_WARNINGS
#include <cstdio>
#include <cctype>
#include <cassert>
#include <cstring>


CDStarRadioHeaderProcessor::CDStarRadioHeaderProcessor() :
m_header(),
m_direct(false),
m_data(false),
m_interrupted(false),
m_control(false),
m_urgent(false),
m_flags(DSRH_NONE),
m_debug(false)
{
	setCallsign(m_ownCallsign1, "", LONG_CALLSIGN_LENGTH);
	setCallsign(m_ownCallsign2, "", SHORT_CALLSIGN_LENGTH);
	setCallsign(m_companionCallsign, "CQCQCQ", LONG_CALLSIGN_LENGTH);
	setCallsign(m_destinationRepeaterCallsign, "DIRECT", LONG_CALLSIGN_LENGTH);
	setCallsign(m_departureRepeaterCallsign, "DIRECT", LONG_CALLSIGN_LENGTH);

	for (unsigned int i = 0U; i < RADIO_HEADER_LENGTH_BITS; i++)
		m_header[i] = false;
}

CDStarRadioHeaderProcessor::~CDStarRadioHeaderProcessor()
{
}

bool CDStarRadioHeaderProcessor::setRadioData(CDStarScrambler& scrambler, const bool* data, unsigned int length)
{
	assert(data != 0);

	if (length != FEC_SECTION_LENGTH_BITS) {
		::fprintf(stderr, "Length of FEC data is incorrect\n");
		return false;
	}

	if (m_debug)
		CUtils::dump("RX Before Descrambling", data, FEC_SECTION_LENGTH_BITS);

	bool buffer1[700U];
	bool buffer2[700U];

	scrambler.process(data, buffer1, FEC_SECTION_LENGTH_BITS);

	if (m_debug)
		CUtils::dump("RX Before Deinterleaving", buffer1, FEC_SECTION_LENGTH_BITS);

	CDStarInterleaver interleaver;
	interleaver.deinterleave(buffer1, buffer2, FEC_SECTION_LENGTH_BITS);

	if (m_debug)
		CUtils::dump("RX Before FEC", buffer2, FEC_SECTION_LENGTH_BITS);

	CDStarFEC fec;
	bool res = fec.decode(buffer2, buffer1, FEC_SECTION_LENGTH_BITS, length);
	if (!res) {
		::fprintf(stderr, "FEC failed to recover the data\n");
		return false;
	}

	if (m_debug)
		CUtils::dump("RX Raw Data", buffer1, RADIO_HEADER_LENGTH_BITS);

	CCCITTChecksum cksum;
	for (unsigned int i = 0U; i < 312U; i += 8U)
		cksum.update(buffer1 + i);

	res = cksum.check(buffer1 + 312U);
	if (!res) {
		::fprintf(stderr, "Radio header checksum failure\n");
		return false;
	}

	for (unsigned int i = 0U; i < RADIO_HEADER_LENGTH_BITS; i++)
		m_header[i] = buffer1[i];

	m_data        = buffer1[0];
	m_direct      = !buffer1[1];
	m_interrupted = buffer1[2];
	m_control     = buffer1[3];
	m_urgent      = buffer1[4];
	m_flags       = DSRH_NONE;

	if (buffer1[5] && buffer1[6] && buffer1[7])
		m_flags = DSRH_REPEATER_CONTROL;
	else if (buffer1[5] && buffer1[6] && !buffer1[7])
		m_flags = DSRH_AUTO_REPLY;
	else if (buffer1[5] && !buffer1[6] && !buffer1[7])
		m_flags = DSRH_REQUEST_RESEND;
	else if (!buffer1[5] && buffer1[6] && buffer1[7])
		m_flags = DSRH_ACK_FLAG;
	else if (!buffer1[5] && buffer1[6] && !buffer1[7])
		m_flags = DSRH_NO_RESPONSE;
	else if (!buffer1[5] && !buffer1[6] && buffer1[7])
		m_flags = DSRH_RELAY_UNAVAILABLE;

	unsigned int n = 24U;
	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++, n += 8U)
		m_destinationRepeaterCallsign[i] = CUtils::bitsToByte(buffer1 + n);
	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++, n += 8U)
		m_departureRepeaterCallsign[i] = CUtils::bitsToByte(buffer1 + n);
	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++, n += 8U)
		m_companionCallsign[i] = CUtils::bitsToByte(buffer1 + n);
	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++, n += 8U)
		m_ownCallsign1[i] = CUtils::bitsToByte(buffer1 + n);
	for (unsigned int i = 0U; i < SHORT_CALLSIGN_LENGTH; i++, n += 8U)
		m_ownCallsign2[i] = CUtils::bitsToByte(buffer1 + n);

	return true;
}

unsigned int CDStarRadioHeaderProcessor::getRadioData(CDStarScrambler& scrambler, bool* data, unsigned int length)
{
	if (length < FEC_SECTION_LENGTH_BITS) {
		::fprintf(stderr, "Buffer is too small for D-Star header\n");
		return 0U;
	}

	bool buffer1[700U];
	bool buffer2[700U];

	for (unsigned int i = 0U; i < RADIO_HEADER_LENGTH_BITS + 6U; i++)
		buffer1[i] = false;

	// Flag 1
	buffer1[0] = m_data;
	buffer1[1] = !m_direct;
	buffer1[2] = m_interrupted;
	buffer1[3] = m_control;
	buffer1[4] = m_urgent;

	switch (m_flags) {
		case DSRH_NONE:
			buffer1[5] = false;
			buffer1[6] = false;
			buffer1[7] = false;
			break;
		case DSRH_RELAY_UNAVAILABLE:
			buffer1[5] = false;
			buffer1[6] = false;
			buffer1[7] = true;
			break;
		case DSRH_NO_RESPONSE:
			buffer1[5] = false;
			buffer1[6] = true;
			buffer1[7] = false;
			break;
		case DSRH_ACK_FLAG:
			buffer1[5] = false;
			buffer1[6] = true;
			buffer1[7] = true;
			break;
		case DSRH_REQUEST_RESEND:
			buffer1[5] = true;
			buffer1[6] = false;
			buffer1[7] = false;
			break;
		case DSRH_AUTO_REPLY:
			buffer1[5] = false;
			buffer1[6] = false;
			buffer1[7] = true;
			break;
		case DSRH_REPEATER_CONTROL:
			buffer1[5] = true;
			buffer1[6] = true;
			buffer1[7] = true;
			break;
	}

	// Flags 2 & 3
	for (unsigned int i = 8U; i < 24U; i++)
		buffer1[i] = false;

	unsigned int n = 24U;
	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++, n += 8U)
		CUtils::byteToBits(m_destinationRepeaterCallsign[i], buffer1 + n);
	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++, n += 8U)
		CUtils::byteToBits(m_departureRepeaterCallsign[i], buffer1 + n);
	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++, n += 8U)
		CUtils::byteToBits(m_companionCallsign[i], buffer1 + n);
	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++, n += 8U)
		CUtils::byteToBits(m_ownCallsign1[i], buffer1 + n);
	for (unsigned int i = 0U; i < SHORT_CALLSIGN_LENGTH; i++, n += 8U)
		CUtils::byteToBits(m_ownCallsign2[i], buffer1 + n);

	CCCITTChecksum cksum;
	for (unsigned int i = 0U; i < 312U; i += 8U)
		cksum.update(buffer1 + i);
	cksum.result(buffer1 + 312U);

	for (unsigned int i = 0U; i < RADIO_HEADER_LENGTH_BITS; i++)
		m_header[i] = buffer1[i];

	if (m_debug)
		CUtils::dump("TX Raw Data", buffer1, 42U * 8U);

	CDStarFEC fec;
	fec.encode(buffer1, buffer2, RADIO_HEADER_LENGTH_BITS, length);

	if (m_debug)
		CUtils::dump("TX After FEC", buffer2, FEC_SECTION_LENGTH_BITS);

	CDStarInterleaver interleaver;
	interleaver.interleave(buffer2, buffer1, FEC_SECTION_LENGTH_BITS);

	if (m_debug)
		CUtils::dump("TX After Interleaving", buffer1, FEC_SECTION_LENGTH_BITS);

	scrambler.reset();
	scrambler.process(buffer1, data, FEC_SECTION_LENGTH_BITS);

	if (m_debug)
		CUtils::dump("TX After Scrambling", data, FEC_SECTION_LENGTH_BITS);

	return FEC_SECTION_LENGTH_BITS;
}

unsigned int CDStarRadioHeaderProcessor::getHeaderData(bool* data, unsigned int length)
{
	if (length < RADIO_HEADER_LENGTH_BITS) {
		::fprintf(stderr, "Length of Radio Header data buffer is too small\n");
		return 0U;
	}

	for (unsigned int i = 0U; i < RADIO_HEADER_LENGTH_BITS; i++)
		data[i] = m_header[i];

	return RADIO_HEADER_LENGTH_BITS;
}

void CDStarRadioHeaderProcessor::setOwnCallsign1(const char* call)
{
	assert(call != 0);

	setCallsign(m_ownCallsign1, call, LONG_CALLSIGN_LENGTH);
}

char* CDStarRadioHeaderProcessor::getOwnCallsign1()
{
	return m_ownCallsign1;
}

void CDStarRadioHeaderProcessor::setOwnCallsign2(const char* call)
{
	assert(call != 0);

	setCallsign(m_ownCallsign2, call, SHORT_CALLSIGN_LENGTH);
}

char* CDStarRadioHeaderProcessor::getOwnCallsign2()
{
	return m_ownCallsign2;
}

void CDStarRadioHeaderProcessor::setCompanionCallsign(const char* call)
{
	assert(call != 0);

	setCallsign(m_companionCallsign, call, LONG_CALLSIGN_LENGTH);
}

char* CDStarRadioHeaderProcessor::getCompanionCallsign()
{
	return m_companionCallsign;
}

void CDStarRadioHeaderProcessor::setDestinationRepeaterCallsign(const char* call)
{
	assert(call != 0);

	setCallsign(m_destinationRepeaterCallsign, call, LONG_CALLSIGN_LENGTH);
}

char* CDStarRadioHeaderProcessor::getDestinationRepeaterCallsign()
{
	return m_destinationRepeaterCallsign;
}

void CDStarRadioHeaderProcessor::setDepartureRepeaterCallsign(const char* call)
{
	assert(call != 0);

	setCallsign(m_departureRepeaterCallsign, call, LONG_CALLSIGN_LENGTH);
}

char* CDStarRadioHeaderProcessor::getDepartureRepeaterCallsign()
{
	return m_departureRepeaterCallsign;
}

void CDStarRadioHeaderProcessor::setDirectMode(bool set)
{
	m_direct = set;

	if (set) {
		setCallsign(m_departureRepeaterCallsign, "DIRECT", LONG_CALLSIGN_LENGTH);
		setCallsign(m_destinationRepeaterCallsign, "DIRECT", LONG_CALLSIGN_LENGTH);
	}
}

bool CDStarRadioHeaderProcessor::isDirectMode() const
{
	return m_direct;
}

void CDStarRadioHeaderProcessor::setDataPacket(bool set)
{
	m_data = set;
}

bool CDStarRadioHeaderProcessor::isDataPacket() const
{
	return m_data;
}

void CDStarRadioHeaderProcessor::setInterrupted(bool set)
{
	m_interrupted = set;
}

bool CDStarRadioHeaderProcessor::isInterrupted() const
{
	return m_interrupted;
}

void CDStarRadioHeaderProcessor::setControlSignal(bool set)
{
	m_control = set;
}

bool CDStarRadioHeaderProcessor::isControlSignal() const
{
	return m_control;
}

void CDStarRadioHeaderProcessor::setUrgent(bool set)
{
	m_urgent = set;
}

bool CDStarRadioHeaderProcessor::isUrgent() const
{
	return m_urgent;
}

void CDStarRadioHeaderProcessor::setRepeaterFlags(DSRH_FLAGS set)
{
	m_flags = set;
}

DSRH_FLAGS CDStarRadioHeaderProcessor::getRepeaterFlags() const
{
	return m_flags;
}

void CDStarRadioHeaderProcessor::setCallsign(char* callOut, const char* callIn, unsigned int length)
{
	size_t callLen = ::strlen(callIn);

	if (callLen >= length) {
		for (unsigned int i = 0U; i < length; i++)
			callOut[i] = ::toupper(callIn[i]);
		callOut[length] = '\0';
	} else {
		for (unsigned int i = 0U; i < callLen; i++)
			callOut[i] = ::toupper(callIn[i]);
		for (unsigned int i = callLen; i < length; i++)
			callOut[i] = ' ';
		callOut[length] = '\0';
	}
}

void CDStarRadioHeaderProcessor::reset()
{
	setCallsign(m_ownCallsign1, "", LONG_CALLSIGN_LENGTH);
	setCallsign(m_ownCallsign2, "", SHORT_CALLSIGN_LENGTH);
	setCallsign(m_companionCallsign, "CQCQCQ", LONG_CALLSIGN_LENGTH);
	setCallsign(m_destinationRepeaterCallsign, "DIRECT", LONG_CALLSIGN_LENGTH);
	setCallsign(m_departureRepeaterCallsign, "DIRECT", LONG_CALLSIGN_LENGTH);

	for (unsigned int i = 0U; i < RADIO_HEADER_LENGTH_BITS; i++)
		m_header[i] = false;

	m_direct = false;
	m_data = false;
	m_interrupted = false;
	m_control = false;
	m_urgent = false;
	m_flags = DSRH_NONE;
}

void CDStarRadioHeaderProcessor::debug(bool set)
{
	m_debug = set;
}
