/*
 *	Copyright (C) 2009 Jonathan Naylor, G4KLX
 *  Copyright (C) 2009 Howard Long, G6LVB
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; version 2 of the License.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 */

#include "GMSKModem.h"

#include <cmath>
#include <cstdio>
#include <cassert>

const float PI = 3.141592654F;

CGMSKModem::CGMSKModem(unsigned int sampleRate, unsigned int symbolRate, float bt) :
m_sampleRate(sampleRate),
m_symbolRate(symbolRate),
m_symbolLen(0U),
m_filter(0),
m_history(0),
m_historyIdx(0U),
m_phase(0U),
m_phaseInc(0U),
m_val(0.0F),
m_bit(false),
m_debug(false)
{
	assert(sampleRate > 0U);
	assert(symbolRate > 0U);
	assert(bt > 0.0F);

	m_symbolLen = sampleRate / symbolRate;
	m_phaseInc  = (unsigned short)(65536U / m_symbolLen);

	m_filter  = new float[m_symbolLen * 4U];
	m_history = new float[m_symbolLen * 4U];

	gaussian(1.0F, bt, m_symbolLen * 4U);
}

CGMSKModem::~CGMSKModem()
{
	delete[] m_filter;
	delete[] m_history;
}

/*
 *	Differential demodulation described in section 4.4.3, page 88,
 *	Modulation and Coding: For Wireless Communications
 *	By Alister Burr
 *	Published by Pearson Education, 2001
 *	ISBN 0201398575, 9780201398571
 *	Also on Google Books
 *
 *	This is slightly modified in that it detects notional bit segregation
 *	by determining the location of the minimum of a[t] * a[t-1] rather than
 *	using the minimum of a[t] * a[t-tsamp].
 *
 *	This might be also be improved by determining bit state by using slope
 *	direction during the half bit time prior to sample time.
 */
MODEM_STATE CGMSKModem::decode(float val)
{
	MODEM_STATE state = MS_UNKNOWN;

	float lastVal = m_val;
	m_val = val;

	bool lastBit = m_bit;

	unsigned short lastPhase = m_phase;
	m_phase += m_phaseInc;

	m_history[m_historyIdx] = m_val * lastVal;

	// Sample at the half bit time...
	if (m_historyIdx > (m_symbolLen / 3U) && m_phase >= 0x8000U && lastPhase <= 0x8000U) {	// Middle of bit?
		m_bit = m_val < 0.0F;

		state = m_bit ? MS_TRUE : MS_FALSE;

		// We should see a large negative peak at zero crossings.
		if (m_bit != lastBit) {	// There was a state change, so let's try to find the point at which it happened
			unsigned int indexMin = 0U;
			float min = 1.0F;	// Maximum positive value

			for (unsigned int i = 0U; i < m_historyIdx; i++) {
				if (m_history[i] < min) {
					indexMin = i;
					min = m_history[i];
				}
			}

			if (indexMin <= m_historyIdx / 2U)
				m_phase += m_phaseInc / 2U;
			else
				m_phase -= m_phaseInc / 2U;
		}

		m_history[0U] = m_history[m_historyIdx];
		m_historyIdx = 1U;
	} else {
		m_historyIdx++;
	}

	return state;
}

unsigned int CGMSKModem::code(bool bit, float* buffer, unsigned int length)
{
	assert(buffer != 0);
	assert(length > 0U);

	if (length < m_symbolLen) {
		::fprintf(stderr, "Output buffer for GMSK data is too small\n");
		return 0U;
	}

	for (unsigned int n = 0U; n < m_symbolLen; n++) {
		// Fill with the new bit
		m_history[m_historyIdx++] = bit ? -0.5F : 0.5F;
		if (m_historyIdx >= (m_symbolLen * 4U))
			m_historyIdx = 0U;

		// Convolve the two
		buffer[n] = 0.0F;
		unsigned int idx = m_historyIdx;
		for (unsigned int i = 0U; i < (m_symbolLen * 4U); i++, idx++) {
			if (idx >= (m_symbolLen * 4U))
				idx = 0U;

			buffer[n] += m_history[idx] * m_filter[i];
		}
	}

	return m_symbolLen;
}

void CGMSKModem::debug(bool set)
{
	m_debug = set;
}

void CGMSKModem::reset()
{
	m_historyIdx = 0U;
	m_phase      = 0U;
	m_phaseInc   = 0U;
	m_val        = 0.0F;
	m_bit        = false;
	m_phaseInc   = (unsigned short)(65536U / m_symbolLen);

	for (unsigned int i = 0U; i < (m_symbolLen * 4U); i++)
		m_history[i] = 0.0F;
}

/*
 *	From GNURadio
 */
void CGMSKModem::gaussian(float gain, float bt, unsigned int taps)
{
	float scale = 0.0F;
	float dt = 1.0F / m_symbolLen;
	float sigma = 1.0F / (::sqrt(::log(2.0F)) / (2.0F * PI * bt));
	float t0 = -0.5F * float(taps);

	for (unsigned int i = 0U; i < taps; i++) {
		t0 += 1.0F;
		float ts = sigma * dt * t0;

		m_filter[i] = ::exp(-0.5F * ts * ts);

		scale += m_filter[i];
	}

	for (unsigned int i = 0U; i < taps; i++)
		m_filter[i] = m_filter[i] / scale * gain;
}
