/*
 *   Copyright (C) 2002-2004,2006-2009 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "WAVFileReader.h"

#include <tchar.h>

#include <cstdio>
#include <cassert>

const int WAVE_FORMAT_IEEE_FLOAT = 3;


CWAVFileReader::CWAVFileReader(TCHAR* fileName) :
m_fileName(NULL),
m_format(FORMAT_16BIT),
m_channels(0U),
m_sampleRate(0U),
m_handle(NULL),
m_parent(),
m_child(),
m_offset(0L)
{
	assert(fileName != NULL);

	m_fileName = ::_tcsdup(fileName);
}

CWAVFileReader::~CWAVFileReader()
{
	::free(m_fileName);
}

bool CWAVFileReader::open()
{
	m_handle = ::mmioOpen(m_fileName, 0, MMIO_READ | MMIO_ALLOCBUF);
	if (m_handle == NULL) {
		::_ftprintf(stderr, _T("WAVFileReader: could not open the WAV file %s\n"), m_fileName);
		return false;
	}

	MMCKINFO parent;
	parent.fccType = mmioFOURCC('W', 'A', 'V', 'E');

	MMRESULT res = ::mmioDescend(m_handle, &parent, 0, MMIO_FINDRIFF);
	if (res != MMSYSERR_NOERROR) {
		::_ftprintf(stderr, _T("WAVFileReader: %s has no \"WAVE\" header\n"), m_fileName);
		return false;
	}

	MMCKINFO child;
	child.ckid = mmioFOURCC('f', 'm', 't', ' ');

	res = ::mmioDescend(m_handle, &child, &parent, MMIO_FINDCHUNK);
	if (res != MMSYSERR_NOERROR) {
		::_ftprintf(stderr, _T("WAVFileReader: %s has no \"fmt \" chunk\n"), m_fileName);
		return false;
	}

	WAVEFORMATEX format;

	LONG len = ::mmioRead(m_handle, (char *)&format, child.cksize);
	if (len != LONG(child.cksize)) {
		::_ftprintf(stderr, _T("WAVFileReader: %s is corrupt, cannot read the WAVEFORMATEX structure\n"), m_fileName);
		return false;
	}
        
	if (format.wFormatTag != WAVE_FORMAT_PCM && format.wFormatTag != WAVE_FORMAT_IEEE_FLOAT) {
		::_ftprintf(stderr, _T("WAVFileReader: %s is not PCM or IEEE Float format, is %u\n"), m_fileName, format.wFormatTag);
		return false;
	}

	m_sampleRate = format.nSamplesPerSec;

	m_channels = format.nChannels;
	if (m_channels > 2U) {
		::_ftprintf(stderr, _T("WAVFileReader: %s has %u channels, more than 2\n"), m_fileName, m_channels);
		return false;
	}

	if (format.wBitsPerSample == 8U && format.wFormatTag == WAVE_FORMAT_PCM) {
		m_format = FORMAT_8BIT;
	} else if (format.wBitsPerSample == 16U && format.wFormatTag == WAVE_FORMAT_PCM) {
		m_format = FORMAT_16BIT;
	} else if (format.wBitsPerSample == 32U && format.wFormatTag == WAVE_FORMAT_IEEE_FLOAT) {
		m_format = FORMAT_32BIT;
	} else {
		::_ftprintf(stderr, _T("WAVFileReader: %s has sample width %u and format %u\n"), m_fileName, format.wBitsPerSample, format.wFormatTag);
		return false;
	}

	res = ::mmioAscend(m_handle, &child, 0);
	if (res != MMSYSERR_NOERROR) {
		::_ftprintf(stderr, _T("WAVFileReader: %s is corrupt, cannot ascend\n"), m_fileName);
		return false;
	}

	child.ckid = mmioFOURCC('d', 'a', 't', 'a');

	res = ::mmioDescend(m_handle, &child, &parent, MMIO_FINDCHUNK);
	if (res != MMSYSERR_NOERROR) {
		::_ftprintf(stderr, _T("WAVFileReader: %s has no \"data\" chunk\n"), m_fileName);
		return false;
	}

	// Get the current location so we can rewind if needed
	m_offset = ::mmioSeek(m_handle, 0L, SEEK_CUR);

	return true;
}

unsigned int CWAVFileReader::read(float* data, unsigned int length)
{
	assert(m_handle != NULL);
	assert(data != NULL);
	assert(length > 0U);

	LONG n = 0L;
	LONG i;

	switch (m_format) {
		case FORMAT_8BIT: {
				unsigned char* buffer8 =  new unsigned char[length * m_channels];

				n = ::mmioRead(m_handle, (char *)buffer8, length * m_channels * sizeof(unsigned char));

				if (n <= 0L) {
					delete[] buffer8;
					return 0U;
				}

				n /= (sizeof(unsigned char) * m_channels);

				switch (m_channels) {
					case 1U:
						for (i = 0L; i < n; i++)
							data[i] = (float(buffer8[i]) - 127.0F) / 128.0F;
						break;
					case 2U:
						for (i = 0L; i < n; i++) {
							data[i * 2U + 0U] = (float(buffer8[i * 2U + 0U]) - 127.0F) / 128.0F;
							data[i * 2U + 1U] = (float(buffer8[i * 2U + 1U]) - 127.0F) / 128.0F;
						}
						break;
				}

				delete[] buffer8;
			}
			break;

		case FORMAT_16BIT: {
				short* buffer16 = new short[length * m_channels];

				n = ::mmioRead(m_handle, (char *)buffer16, length * m_channels * sizeof(short));

				if (n <= 0L) {
					delete[] buffer16;
					return 0U;
				}

				n /= (sizeof(short) * m_channels);

				switch (m_channels) {
					case 1U:
						for (i = 0L; i < n; i++)
							data[i] = float(buffer16[i]) / 32768.0F;
						break;
					case 2U:
						for (i = 0L; i < n; i++) {
							data[i * 2U + 0U] = float(buffer16[i * 2U + 0U]) / 32768.0F;
							data[i * 2U + 1U] = float(buffer16[i * 2U + 1U]) / 32768.0F;
						}
						break;
				}

				delete[] buffer16;
			}
			break;

		case FORMAT_32BIT: {
				float* buffer32 = new float[length * m_channels];

				n = ::mmioRead(m_handle, (char *)buffer32, length * m_channels * sizeof(float));

				if (n <= 0L) {
					delete[] buffer32;
					return 0U;
				}

				n /= (sizeof(float) * m_channels);

				switch (m_channels) {
					case 1U:
						for (i = 0L; i < n; i++)
							data[i] = buffer32[i];
						break;
					case 2U:
						// Swap I and Q for SDR-1000 data
						for (i = 0L; i < n; i++) {
							data[i * 2U + 0U] = buffer32[i * 2U + 1U];
							data[i * 2U + 1U] = buffer32[i * 2U + 0U];
						}
						break;
				}

				delete[] buffer32;
			}
			break;
	}

	return n;
}

void CWAVFileReader::rewind()
{
	assert(m_handle != NULL);

	::mmioSeek(m_handle, m_offset, SEEK_SET);
}

void CWAVFileReader::close()
{
	assert(m_handle != 0);

	::mmioClose(m_handle, 0U);

	m_handle = NULL;
}

unsigned int CWAVFileReader::getSampleRate() const
{
	return m_sampleRate;
}

unsigned int CWAVFileReader::getChannels() const
{
	return m_channels;
}

