/*
 *   Copyright (C) 2002-2004,2006-2009 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "WAVFileWriter.h"

#include <cstdio>
#include <cassert>

const int WAVE_FORMAT_IEEE_FLOAT = 3;

CWAVFileWriter::CWAVFileWriter(TCHAR* fileName, unsigned int sampleRate, unsigned int channels, unsigned int sampleWidth) :
m_fileName(NULL),
m_sampleRate(sampleRate),
m_channels(channels),
m_sampleWidth(sampleWidth),
m_handle(NULL),
m_parent(),
m_child()
{
	assert(fileName != NULL);
	assert(sampleRate > 0U);
	assert(channels == 1U || channels == 2U);
	assert(sampleWidth == 8U || sampleWidth == 16U || sampleWidth == 32U);

	m_fileName = ::_tcsdup(fileName);
}

CWAVFileWriter::~CWAVFileWriter()
{
	::free(m_fileName);
}

bool CWAVFileWriter::open()
{
	m_handle = ::mmioOpen(m_fileName, 0, MMIO_WRITE | MMIO_CREATE | MMIO_ALLOCBUF);
	if (m_handle == NULL) {
		::_ftprintf(stderr, _T("WAVFileWriter: could not open the file %s\n"), m_fileName);
		return false;
	}

	m_parent.fccType = mmioFOURCC('W', 'A', 'V', 'E');
	m_parent.cksize  = 0;

	MMRESULT res = ::mmioCreateChunk(m_handle, &m_parent, MMIO_CREATERIFF);
	if (res != MMSYSERR_NOERROR) {
		::_ftprintf(stderr, _T("WAVFileWriter: could not write to file %s\n"), m_fileName);
		return false;
	}

	m_child.ckid   = mmioFOURCC('f', 'm', 't', ' ');
	m_child.cksize = sizeof(WAVEFORMATEX);

	res = ::mmioCreateChunk(m_handle, &m_child, 0);
	if (res != MMSYSERR_NOERROR) {
		::_ftprintf(stderr, _T("WAVFileWriter: could not write to the file %s\n"), m_fileName);
		return false;
	}

	WAVEFORMATEX format;
	format.wBitsPerSample  = m_sampleWidth;
	if (m_sampleWidth == 8U || m_sampleWidth == 16U)
		format.wFormatTag = WAVE_FORMAT_PCM;
	else
		format.wFormatTag = WAVE_FORMAT_IEEE_FLOAT;
	format.nChannels       = m_channels;
	format.nSamplesPerSec  = m_sampleRate;
	format.nAvgBytesPerSec = m_sampleRate * m_channels * m_sampleWidth / 8;
	format.nBlockAlign     = m_channels * m_sampleWidth / 8;
	format.cbSize          = 0;

	LONG n = ::mmioWrite(m_handle, (CHAR *)&format, sizeof(WAVEFORMATEX));
	if (n != sizeof(WAVEFORMATEX)) {
		::_ftprintf(stderr, _T("WAVFileWriter: could not write to the file %s\n"), m_fileName);
		return false;
	}

	::mmioAscend(m_handle, &m_child, 0);

	m_child.ckid   = mmioFOURCC('d', 'a', 't', 'a');
	m_child.cksize = 0;

	res = ::mmioCreateChunk(m_handle, &m_child, 0);
	if (res != MMSYSERR_NOERROR) {
		::_ftprintf(stderr, _T("WAVFileWriter: could not write to the file %s\n"), m_fileName);
		return false;
	}

	return true;
}

bool CWAVFileWriter::write(const float* buffer, unsigned int length)
{
	assert(m_handle != NULL);
	assert(buffer != NULL);
	assert(length > 0U);

	unsigned int i;
	LONG bytes = 0L;
	LONG n = 0L;

	switch (m_sampleWidth) {
		case 8U: {
				unsigned char* buffer8 = new unsigned char[length * m_channels];

				switch (m_channels) {
					case 1U:
						for (i = 0U; i < length; i++)
							buffer8[i] = (unsigned char)(buffer[i] * 128.0F + 127.0F);
						break;
					case 2U:
						for (i = 0U; i < (length * 2U); i++)
							buffer8[i] = (unsigned char)(buffer[i] * 128.0F + 127.0F);
						break;
				}

				bytes = length * m_channels * sizeof(unsigned char);

				n = ::mmioWrite(m_handle, (char *)buffer8, bytes);

				delete[] buffer8;
			}
			break;

		case 16U: {
				short* buffer16 = new short[length * m_channels];

				switch (m_channels) {
					case 1U:
						for (i = 0U; i < length; i++)
							buffer16[i] = short(buffer[i] * 32768.0F);
						break;
					case 2U:
						for (i = 0U; i < (length * 2U); i++)
							buffer16[i] = short(buffer[i] * 32768.0F);
						break;
				}

				bytes = length * m_channels * sizeof(short);

				n = ::mmioWrite(m_handle, (char *)buffer16, bytes);

				delete[] buffer16;
			}
			break;

		case 32U: {
				float* buffer32 = new float[length * m_channels];

				switch (m_channels) {
					case 1U:
						for (i = 0U; i < length; i++)
							buffer32[i] = float(buffer[i]);
						break;
					case 2U:
						// Swap I and Q
						for (i = 0U; i < length; i++) {
							buffer32[i * 2U + 0U] = float(buffer[i * 2U + 1U]);
							buffer32[i * 2U + 1U] = float(buffer[i * 2U + 0U]);
						}
						break;
				}

				bytes = length * m_channels * sizeof(float);

				n = ::mmioWrite(m_handle, (char *)buffer32, bytes);

				delete[] buffer32;
			}
			break;
	}

	return n == bytes;
}

void CWAVFileWriter::close()
{
	assert(m_handle != NULL);

	::mmioAscend(m_handle, &m_child, 0);
	::mmioAscend(m_handle, &m_parent, 0);

	::mmioClose(m_handle, 0);
	m_handle = NULL;
}

unsigned int CWAVFileWriter::getSampleRate() const
{
	return m_sampleRate;
}

unsigned int CWAVFileWriter::getChannels() const
{
	return m_channels;
}
