/*
 *	Copyright (C) 2009 Jonathan Naylor, G4KLX
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; version 2 of the License.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 */

#include "DStarTransmitter.h"

#include "WAVFileWriter.h"
#include "DStarDefines.h"
#include "Utils.h"

#include <windows.h>
#include <tchar.h>

#include <cassert>
#include <ctime>

const unsigned int SAMPLE_RATE = 48000U;

const unsigned int FILENAME_LENGTH = 50U;
const unsigned int BUFFER_LENGTH   = 256U;

int _tmain(int argc, _TCHAR* argv[])
{
	if (argc < 3) {
		::_ftprintf(stderr, _T("Usage: DStarTransmitter <Info File> <AMBE Data File>\n"));
		return 1;
	}

	FILE* fpAMBE = ::_tfopen(argv[2], _T("rb"));
	if (fpAMBE == NULL) {
		::_ftprintf(stderr, _T("Cannot open the AMBE data file for reading - %s\n"), argv[2]);
		return 1;
	}

	FILE* fpInfo = ::_tfopen(argv[1], _T("rt"));
	if (fpInfo == NULL) {
		::_ftprintf(stderr, _T("Cannot open the INFO data file for reading - %s\n"), argv[1]);
		::fclose(fpAMBE);
		return 1;
	}

	time_t now = ::time(NULL);
	struct tm* tm = ::localtime(&now);

	TCHAR fileName[FILENAME_LENGTH];
	::_tcsftime(fileName, FILENAME_LENGTH, _T("DStar_%Y%m%d_%H%M%S.wav"), tm);

	::_ftprintf(stdout, _T("Writing WAV data to %s\n"), fileName);

	CWAVFileWriter writer(fileName, SAMPLE_RATE, 1U, 16U);
	bool res = writer.open();
	if (!res) {
		::fclose(fpAMBE);
		::fclose(fpInfo);
		return 1;
	}

	CDStarTransmitter transmitter(&writer, fpInfo, fpAMBE);
	res = transmitter.run();
	if (!res) {
		writer.close();
		::fclose(fpAMBE);
		::fclose(fpInfo);
		return 1;
	}

	writer.close();
	::fclose(fpInfo);
	::fclose(fpAMBE);

	return 0;
}

CDStarTransmitter::CDStarTransmitter(ISoundWriter *writer, FILE *fpInfo, FILE *fpAMBE) :
m_header(),
m_scrambler(),
m_modem(SAMPLE_RATE, GMSK_BIT_RATE, GMSK_BT),
m_writer(writer),
m_fpInfo(fpInfo),
m_fpAMBE(fpAMBE),
m_buffer(),
m_frameNo(0U),
m_bitLen(0U),
m_debug(false)
{
	assert(writer != NULL);
	assert(fpInfo != NULL);
	assert(fpAMBE != NULL);

	m_bitLen = m_writer->getSampleRate() / GMSK_BIT_RATE;

	m_buffer = new float[m_bitLen];
}

CDStarTransmitter::~CDStarTransmitter()
{
	delete[] m_buffer;
}

bool CDStarTransmitter::run()
{
	bool res = writeBitSync();
	if (!res)
		return false;

	::fprintf(stdout, "Written the bit sync\n");

	res = writeFrameSync();
	if (!res)
		return false;

	::fprintf(stdout, "Written the frame sync\n");

	res = parseInfoFile();
	if (!res)
		return false;

	::fprintf(stdout, "Parsed the info file\n");

	res = writeRadioHeader();
	if (!res)
		return false;

	::fprintf(stdout, "Written the radio header\n");

	res = writeAudio();
	if (!res)
		return false;

	res = writeEndPattern();

	return res;
}

bool CDStarTransmitter::writeBitSync()
{
	for (unsigned int i = 0U; i < 100U; i++) {
		bool res = writeBits(GMSK_BIT_SYNC, 4U);
		if (!res)
			return false;
	}

	return true;
}

bool CDStarTransmitter::writeFrameSync()
{
	return writeBits(FRAME_SYNC, 15U);
}

bool CDStarTransmitter::writeRadioHeader()
{
	m_scrambler.reset();

	bool data[FEC_SECTION_LENGTH_BITS];

	unsigned int length = m_header.getRadioData(m_scrambler, data, FEC_SECTION_LENGTH_BITS);

	return writeBits(data, length);
}

bool CDStarTransmitter::writeAudio()
{
	m_frameNo = 0U;

	// Get the raw header data, no scrambling, interleaving, or FEC
	bool header[RADIO_HEADER_LENGTH_BITS];
	unsigned int n = m_header.getHeaderData(header, RADIO_HEADER_LENGTH_BITS);
	if (n == 0U)
		return false;

	// Add the type and length headers and filler up to 60 bytes
	unsigned int bits = 0U;
	unsigned int bytes = 0U;
	bool slowData[SLOW_DATA_LENGTH_BITS];
	for (unsigned int i = 0U; i < SLOW_DATA_LENGTH_BITS; i += 8U) {
		if (bytes < RADIO_HEADER_LENGTH_BYTES) {
			unsigned int remain = RADIO_HEADER_LENGTH_BYTES - bytes;

			if ((i % 48U) == 0U) {
				if (remain >= 5U)
					CUtils::byteToBitsRev(SLOW_DATA_TYPE | 5U, slowData + i);
				else
					CUtils::byteToBitsRev(SLOW_DATA_TYPE | remain, slowData + i);
			} else {
				for (int j = 7U; j >= 0; j--)
					slowData[i + j] = header[bits++];
				bytes++;
			}
		} else {
			CUtils::byteToBitsRev('f', slowData + i);		// Pad with f's until the end
		}
	}

	unsigned int  nData = 0U;
	unsigned char audio[VOICE_FRAME_LENGTH_BYTES];

	size_t audioLen;
	while ((audioLen = ::fread(audio, 1U, VOICE_FRAME_LENGTH_BYTES, m_fpAMBE)) > 0U) {
		bool frame[DV_FRAME_LENGTH_BITS];

		for (unsigned int i = 0U; i < DV_FRAME_LENGTH_BITS; i++)
			frame[i] = false;

		// Always copy the audio data
		unsigned int n = 0U;
		for (unsigned int i = 0U; i < audioLen; i++, n += 8U)
			CUtils::byteToBitsRev(audio[i], frame + n);

		// Is this a re-sync frame?
		if ((m_frameNo % 21U) == 0U) {
			::fprintf(stdout, "Writing a data re-sync in frame %u\n", m_frameNo);

			for (unsigned int i = 0U; i < DATA_FRAME_LENGTH_BITS; i++)
				frame[i + VOICE_FRAME_LENGTH_BITS] = DATA_SYNC[i];
		} else {
			for (unsigned int i = 0U; i < DATA_FRAME_LENGTH_BITS; i++) {
				frame[i + VOICE_FRAME_LENGTH_BITS] = slowData[nData++];
				if (nData >= SLOW_DATA_LENGTH_BITS)
					nData = 0U;
			}

			m_scrambler.reset();
			m_scrambler.process(frame + VOICE_FRAME_LENGTH_BITS, DATA_FRAME_LENGTH_BITS);
		}

		bool res = writeBits(frame, DV_FRAME_LENGTH_BITS);
		if (!res)
			return false;

		m_frameNo++;
	}

	return true;
}

bool CDStarTransmitter::writeEndPattern()
{
	::fprintf(stdout, "Writing end pattern in frame %u\n", m_frameNo);

	return writeBits(END_PATTERN, END_PATTERN_LENGTH);
}

bool CDStarTransmitter::writeBits(const bool* data, unsigned int length)
{
	assert(data != NULL);
	assert(length > 0U);

	for (unsigned int i = 0U; i < length; i++) {
		unsigned int n = m_modem.code(data[i], m_buffer, m_bitLen);

		bool res = m_writer->write(m_buffer, n);
		if (!res)
			return false;
	}

	return true;
}

bool CDStarTransmitter::parseInfoFile()
{
	char buffer[BUFFER_LENGTH];

	while (::fgets(buffer, BUFFER_LENGTH, m_fpInfo) != NULL) {
		char* s = strchr(buffer, '\n');
		if (s != NULL) *s = '\0';
		s = strchr(buffer, '\r');
		if (s != NULL) *s = '\0';

		if (::strcmp(buffer, "Type: Data packet") == 0)
			m_header.setDataPacket(true);
		else if (::strcmp(buffer, "Type: Voice packet") == 0)
			m_header.setDataPacket(false);
		else if (::strcmp(buffer, "Mode: Direct connection") == 0)
			m_header.setDirectMode(true);
		else if (::strcmp(buffer, "Mode: Via repeater") == 0)
			m_header.setDirectMode(false);
		else if (::strcmp(buffer, "Interrupted: Yes") == 0)
			m_header.setInterrupted(true);
		else if (::strcmp(buffer, "Interrupted: No") == 0)
			m_header.setInterrupted(false);
		else if (::strcmp(buffer, "Control signal: Yes") == 0)
			m_header.setControlSignal(true);
		else if (::strcmp(buffer, "Control signal: No") == 0)
			m_header.setControlSignal(false);
		else if (::strcmp(buffer, "Urgent priority: Yes") == 0)
			m_header.setUrgent(true);
		else if (::strcmp(buffer, "Urgent priority: No") == 0)
			m_header.setUrgent(false);
		else if (::strcmp(buffer, "Repeater flags: Control mode") == 0)
			m_header.setRepeaterFlags(DSRH_REPEATER_CONTROL);
		else if (::strcmp(buffer, "Repeater flags: Auto reply") == 0)
			m_header.setRepeaterFlags(DSRH_AUTO_REPLY);
		else if (::strcmp(buffer, "Repeater flags: Request resent") == 0)
			m_header.setRepeaterFlags(DSRH_REQUEST_RESEND);
		else if (::strcmp(buffer, "Repeater flags: ACK flag") == 0)
			m_header.setRepeaterFlags(DSRH_ACK_FLAG);
		else if (::strcmp(buffer, "Repeater flags: No response available") == 0)
			m_header.setRepeaterFlags(DSRH_NO_RESPONSE);
		else if (::strcmp(buffer, "Repeater flags: Relay unavailable") == 0)
			m_header.setRepeaterFlags(DSRH_RELAY_UNAVAILABLE);
		else if (::strcmp(buffer, "Repeater flags: None") == 0)
			m_header.setRepeaterFlags(DSRH_NONE);
		else if (::strncmp(buffer, "Destination repeater callsign: ", 31U) == 0)
			m_header.setDestinationRepeaterCallsign(buffer + 31U);
		else if (::strncmp(buffer, "Departure repeater callsign: ", 29U) == 0)
			m_header.setDepartureRepeaterCallsign(buffer + 29U);
		else if (::strncmp(buffer, "Companion callsign: ", 20U) == 0)
			m_header.setCompanionCallsign(buffer + 20U);
		else if (::strncmp(buffer, "Own callsign 1: ", 16U) == 0)
			m_header.setOwnCallsign1(buffer + 16U);
		else if (::strncmp(buffer, "Own callsign 2: ", 16U) == 0)
			m_header.setOwnCallsign2(buffer + 16U);
	}

	return true;
}

void CDStarTransmitter::debug(bool set)
{
	m_debug = set;

	m_modem.debug(set);
	m_header.debug(set);
}
